#include "config.h"

#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>

#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>

/* Weird linux bug? */
#ifdef __PRI64_PREFIX
#undef __PRI64_PREFIX
#define __PRI64_PREFIX "l"
#endif

#include <regex.h>
#include <ctype.h>
#include <fcntl.h>

#include <iostream>

#include <tsk3/libtsk.h>
#include "fiwalk.h"
#include "content.h"
#include "plugin.h"

#ifdef HAVE_MAGIC_H
#include <magic.h>
#endif

#include <err.h>

/****************************************************************
 ** Content Output
 ****************************************************************/

/** set the filename and see if we need to do the plugin
 */
void content::set_filename(const string &filename)
{
    this->evidence_filename = filename;
    this->do_plugin = plugin_match(this->evidence_filename);
    if(do_plugin || opt_magic) open_tempfile();
    if(opt_save)               open_savefile();
}


bool content::name_filtered()
{
    if(namelist.size()>0 && evidence_filename.size()>0){
	for(vector<string>::const_iterator i = namelist.begin(); i!=namelist.end(); i++){
	    if(strcasestr(evidence_filename.c_str(),i->c_str())){
		return false;		// no, name is not filtered
	    }
	}
	return true;			// name is filtered because it was not wanted
    }
    return false;			// name is not filtered because there is no filter list
}



/** Open a temporary file to put the file for the plugin and/or the file command
 */
void content::open_tempfile()
{
    if(tempfile) return;		//  already opened

    int added = 0;
    tempfile_path = tempdir + "/";
    for(string::const_iterator cc = evidence_filename.begin();
	cc != evidence_filename.end();
	cc++){
	if(isalnum(*cc)){
	    tempfile_path.push_back(*cc);
	    added++;
	}
    }
    if(added==0) tempfile_path += "tempfile";

    tempfile = fopen(tempfile_path.c_str(),"w");
    if(!tempfile) errx(1,"%s:",tempfile_path.c_str());
}


/** Open a file where a bytestream will be saved.
 *
 * @param filename - specifies the name of the file. If the file
 * exists, then start a counter between the end of the file and the
 * extension and keep incrementing...
 *
 */
void content::open_savefile()
{
    /* Figure out the pathname to save */
    save_path = save_outdir + "/" + evidence_filename;

    /* IF we can open this directory directly, just take it */
    if(access(save_path.c_str(),F_OK)!=0){
	savefile = fopen(save_path.c_str(),"w");
	return;
    }

    /* Make a format string that is filename.%d.ext
     */

    string fmtstring = save_path;

    int ext = fmtstring.rfind('.');
    if(ext==-1){
	fmtstring += ".%03d";
    }
    else{
	fmtstring.insert(ext,".%03d.");
    }

    for(int i=1;i<10000;i++){
	char fname[1024];
	snprintf(fname,sizeof(fname),fmtstring.c_str(),i);
	if(access(fname,F_OK)!=0){
	    savefile = fopen(fname,"w");
	    return;
	}
    }
    // hm. Can't open file?
    err(1,"Cannot open '%s'",save_path.c_str());
}

/**
 * run the file command.
 * -b = do not put the filename in the output.
 * -z = attempt to decompress compressed files.
 */
 
string content::filemagic()
{
#ifdef HAVE_LIBMAGIC
    static bool magic_init=false;
    static bool magic_bad = false;
    static magic_t mt;

    if(magic_bad) return string("");

    if(magic_init==false){
	magic_init=true;
	mt = magic_open(MAGIC_NONE);
	if(magic_load(mt,NULL)==-1){
	    magic_bad = true;
	    return string("");
	}
    }
    const char *ret = magic_file(mt,tempfile_path.c_str());
    return string(ret ? ret : "");
#else
    char cmd[1024];
    char buf[1024];
    string out;
    snprintf(cmd,sizeof(cmd),"file -b -z '%s'",tempfile_path.c_str());
    FILE *f = popen(cmd,"r");
    while(!feof(f)){
	if(fgets(buf,sizeof(buf),f)) out += buf;
    }
    pclose(f);
    while(out.size()>0 && out[out.size()-1]=='\n'){
	out.erase(out.size()-1);
    }
    return out;
#endif
}


void content::write_record()
{
    if(opt_magic)		file_info("libmagic",this->filemagic());
    if(this->segs.size()>0){
	string runz = "";
	for(seglist::const_iterator i = this->segs.begin();i!=this->segs.end();i++){
	    char buf[1024];
	    if(i->flags & TSK_FS_BLOCK_FLAG_SPARSE){
		sprintf(buf,"<run file_offset='%"PRIu64"' fill='0' len='%"PRIu64"'/>",i->file_offset,i->len);
	    } else if (i->flags & TSK_FS_BLOCK_FLAG_RAW){
		sprintf(buf,"<run file_offset='%"PRIu64"' fs_offset='%"PRIu64"' img_offset='%"PRIu64"' len='%"PRIu64"'/>",
			i->file_offset,i->fs_offset,i->img_offset,i->len);
	    } else if (i->flags & TSK_FS_BLOCK_FLAG_COMP){
		if(i->fs_offset){
		    sprintf(buf,"<run file_offset='%"PRIu64"' fs_offset='%"PRIu64"' img_offset='%"PRIu64"' uncompressed_len='%"PRIu64"'/>",
			    i->file_offset,i->fs_offset,i->img_offset,i->len);
		} else {
		    sprintf(buf,"<run file_offset='%"PRIu64"' uncompressed_len='%"PRIu64"'/>",i->file_offset,i->len);
		}
	    } else{
		sprintf(buf,"<run file_offset='%"PRIu64"' unknown_flags='%d'/>",i->file_offset,i->flags);
	    }
	    runz += buf;
	}
	file_info_xml("byte_runs",runz);
	if(opt_md5)  file_info("md5", h_md5.hexhash());
	if(opt_sha1) file_info("sha1",h_sha1.hexhash());
    }

    /* This stuff is only if we are creating ARFF output */
    if(a){
	file_info("fragments",this->segs.size());
	if(sector_size>0){
	    if(this->segs.size()>=1){
		int64_t frag1start = this->segs[0].img_offset / sector_size;
		file_info("frag1startsector",frag1start);
	    }
	    if(this->segs.size()>=2){
		int64_t frag2start = this->segs[1].img_offset / sector_size;
		file_info("frag2startsector",frag2start);
	    }
	}
    }
    //if(opt_userguess && this->has_filename()) add_userguess(this->filename());
    //if(this->do_plugin) plugin_process(this->tempfile_path);
}

/* Do we need full content? */
bool content::need_file_walk()
{
    return opt_md5 || opt_sha1 || opt_save || do_plugin  || (mybloomset.size()>0) || opt_magic || opt_get_fragments; 
}

void content::add_seg(int64_t img_offset,int64_t fs_offset,int64_t file_offset,int64_t len,TSK_FS_BLOCK_FLAG_ENUM flags)
{
    struct seg newseg;
    newseg.img_offset = img_offset;
    newseg.fs_offset = fs_offset;
    newseg.file_offset = file_offset;
    newseg.len   = len;
    newseg.flags = flags;
    this->segs.push_back(newseg);
}

void content::add_bytes(const char *buf,size_t size)
{
    if(opt_md5)   h_md5.update((unsigned char *)buf,size);
    if(opt_sha1)  h_sha1.update((unsigned char *)buf,size);
    if(savefile){
	if(fwrite(buf,size,1,savefile)!=1) errx(1,"fwrite");
    }
    if(tempfile){
	if(fwrite(buf,size,1,tempfile)!=1) errx(1,"fwrite");
    }
}


void 	content::final()
{
    if(savefile){ fclose(savefile); savefile = 0; }
    if(tempfile){ fclose(tempfile); tempfile = 0; }
}
